/* 
 * The Finite State Machine Explorer
 * Copyright (c) 1995-1996
 * Matt Chapman (csuoq@csv.warwick.ac.uk)
 * Dept. Computer Science
 * University of Warwick, Coventry, UK.
 */

package Machine;

/**
 * A class to represent a single connection in a finite state machine
 * @version    19 Jan 1996
 * @author     Matt Chapman
 */
public final class Connection extends Object 
{
    // each connection is between two states
    private State fromState,toState;

    // the transition characters for the connection
    private String trans;

    // the corners of the transition box
    private Coord topleft;
    private Coord botright;

    // coordinates of the connection arrow
    public Coord arrowtip;
    public Coord arrowedge1;
    public Coord arrowedge2;

    // constants for the size of the arrow and width of text box
    private final float ARROWL = 0.005f;
    private final float TWIDTH = 0.008f;

    // precomputed constant, for efficiency
    private static final float sq3over2 = (float)(Math.sqrt(3)/2);

    // the current width and height of the transition details
    private float width = TWIDTH;
    private float height = 0.01f;

    /**
     * Creates a connection entry.
     * @param from    the state the connection starts from
     * @param to      the state the connection goes to
     * @param t       the transition character
     */
    public Connection(State from, State to, String t) 
    {
		  fromState = from;
		  toState = to;
		  trans = new String(t);
    }

    /**
     * Calculates a position for the top left corner of the box containing the
     * transition details, based on the states it joins.
     */
    public void calculatePos() 
    {
		  float midx,midy;

		  // connection is to and from same state
		  if (fromState == toState) 
		  {
				// left top of transition box
				midx = fromState.getPos().x + 1.8f*fromState.radius;
				midy = fromState.getPos().y + 1.8f*fromState.radius - height/2.0f;

				float xi = fromState.getPos().x;
				float yi = fromState.getPos().y + fromState.radius;
				float dx = fromState.radius/2;
				float dy = fromState.radius * sq3over2;

				arrowtip = new Coord(xi,yi);
				Coord arrowbase = new Coord(xi+dx,yi+dy);
				float xoff = sq3over2 * ARROWL;
				float yoff = 0.5f * ARROWL;
				arrowedge1 = new Coord(arrowbase.x + xoff,arrowbase.y - yoff);
				arrowedge2 = new Coord(arrowbase.x - xoff,arrowbase.y + yoff);

		  }
		  else { // connection is between two distinct states
				float xdiff = (toState.getPos().x - fromState.getPos().x);
				float ydiff = (toState.getPos().y - fromState.getPos().y);

				// calculate arrow position
				double theta = Math.atan(ydiff/xdiff);
				float xoff = (float)(Math.sin(theta) * ARROWL);
				float yoff = (float)(Math.cos(theta) * ARROWL);

				float bx = (float)(Math.sin(theta) * TWIDTH);
				float by = (float)(Math.cos(theta) * height);

				// left top of transition box
				midx = xdiff * 0.58f + fromState.getPos().x - TWIDTH/2 + bx;
				midy = ydiff * 0.58f + fromState.getPos().y - height/2 - by;

				float edgex = (float)(fromState.radius * 0.9 * Math.cos(theta));
				float edgey = (float)(fromState.radius * 0.9 * Math.sin(theta));
				if (xdiff < 0.0f) 
				{
					 edgex = -edgex;
					 edgey = -edgey;
				}
				arrowtip = new Coord(toState.getPos().x - edgex,
											toState.getPos().y - edgey);
				Coord arrowbase = new Coord(toState.getPos().x - edgex*2,
													 toState.getPos().y - edgey*2);
				arrowedge1 = new Coord(arrowbase.x + xoff,arrowbase.y - yoff);
				arrowedge2 = new Coord(arrowbase.x - xoff,arrowbase.y + yoff);
		  }
		  topleft = new Coord(midx,midy);
		  botright = new Coord(midx+width,midy+height);
    }

    /**
     * Gets the positon of the top left corner of the box containing the
     * transition details.
     * @return the coordinate of the connection
     */
    public Coord getTopLeft() 
    {
		  return topleft;
    }

    /**
     * Gets the positon of the bottom right corner of the box containing the
     * transition details.
     * @return the coordinate of the connection
     */
    public Coord getBotRight() 
    {
		  return botright;
    }

    /**
     * Gets the state of the start of the connection.
     * @return the state
     */
    public State getFrom() 
    {
		  return fromState;
    }

    /**
     * Gets the transition string of the connection.
     * @return the string
     */
    public String getTrans() 
    {
		  return trans;
    }

    /**
     * Sets the transition string of the connection.
     * @param n the new transition string
     */
    public void setTrans(String n) 
    {
		  trans = n;
		  width = n.length() * TWIDTH;
    }

    /**
     * Gets the state of the end of the connection.
     * @return the state
     */
    public State getTo() 
    {
		  return toState;
    }
}
