/* 
 * The Finite State Machine Explorer
 * Copyright (c) 1995-1996
 * Matt Chapman (csuoq@csv.warwick.ac.uk)
 * Dept. Computer Science
 * University of Warwick, Coventry, UK.
 */

package Machine;

/**
 * A class to represent perform conversion from a NFA to a DFA
 * @version    12 March 1996
 * @author     Matt Chapman
 */
public final class Convert 
{

    /**
     * Find the transition alphabet of the given machine
     * @param m the machine of interest
     * @return a set containing all the transition characters used
     */
    public Set transitionAlph(Logical m) 
    {
		  Set alph = new Set();
		  // examine every connection
		  for (int i=0; i<m.numConnections(); i++) 
		  {
				Connection c = m.connectionAt(i);
				String str = c.getTrans();
				for (int j=0; j<str.length(); j++) 
				{
					 Character t = new Character(str.charAt(j));
					 alph.addElement(t);
				}
		  }
		  return alph;
    }

    /**
     * Converts a non-deterministic machine to a deterministic one.
     * The resulting machine is then minimized. The number of states can
     * increase from N to 2^N, so for large machine the process is going
     * to be very slow.
     * @param m the machine to be converted
     * @return the converted machine
     */
    public Logical NFA2DFA(Logical m) 
    {
		  // a new machine, in which the conversion is constructed
		  Logical dfa = new Logical();

		  // get transition alphabet
		  Set ta = transitionAlph(m);

		  // create a set of all the states in the machine, and a set T of all
		  // accepting states
		  Set ostates = new Set();
		  Set T = new Set();
		  for (int i=0; i<m.numStates(); i++) 
		  {
				State s = m.stateAt(i);
				ostates.addElement(s);
				if (s.isAccepting())
					 T.addElement(s);
		  }

		  // the states of the new machine are the powerset of the states of
		  // the old machine
		  Set nstates = ostates.powerSet();

		  // an array of the sets of the new machine
		  Set tf[] = new Set[nstates.size()];

		  // an array of the states of the new machine
		  State dfaState[] = new State[tf.length];

		  // create the states of the new machine, using existing states where
		  // possible
		  for (int i=0; i<nstates.size(); i++) 
		  {
				tf[i] = (Set)nstates.elementAt(i);
				if (tf[i].size() == 1) 
				{
					 dfaState[i] = (State)tf[i].elementAt(0);
					 dfa.addGiven(dfaState[i]);
				}
				else 
				{
					 dfaState[i] = dfa.addState();
				}
				if (dfaState[i] == m.getStart())
					 dfa.makeStart(dfaState[i]);

				// make state accepting if intersection with T is non-empty
				if (tf[i].intersection(T).size() > 0)
					 dfaState[i].makeAccepting();
		  }

		  // generate transition function for all states of new machine
		  for (int i=0; i<tf.length; i++) 
		  {
				// for each element in the transition alphabet
				for (int j=0; j<ta.size(); j++) 
				{
					 Character ct = (Character)ta.elementAt(j);
					 char t = ct.charValue();
					 Set trans = new Set();
					 for (int k=0; k<tf[i].size(); k++) 
					 {
						  trans.union(m.transitionSet((State)tf[i].elementAt(k),
																t));
					 }
					 // find transtion set in list, to identify the required state
					 int index=-1;
					 for (int l=0; l<tf.length; l++) 
					 {
						  if (trans.equals(tf[l]))
								index=l;
					 }
					 // add appropriate connection to machine
					 dfa.addConnection(dfaState[i],dfaState[index],t);
				}
		  }
		  return dfa;
    }

    /**
     * Determine whether the given machine is deterministic
     * @param m the machine of interest
     * @return true if m is deterministic, false otherwise
     */
    public boolean isDeterministic(Logical m)
    {
		  Set ta = transitionAlph(m);
		  for (int i=0; i<m.numStates(); i++)
		  {
				State s = m.stateAt(i);
				// see if more than one connection leaves this state for a given
				// transition character
				for (int j=0; j<ta.size(); j++)
				{
					 Character ct = (Character)ta.elementAt(j);
					 char t = ct.charValue();
					 if (m.transitionSet(s,t).size() > 1)
						  return false;
				}
		  }
		  return true;
    }

}
