/* 
 * The Finite State Machine Explorer
 * Copyright (c) 1995-1996
 * Matt Chapman (csuoq@csv.warwick.ac.uk)
 * Dept. Computer Science
 * University of Warwick, Coventry, UK.
 */

package Machine;

/**
 * A class to evaluate the layout of a machine according to various criteria
 * @version 10 Feb 1996
 * @author  Matt Chapman
 */
public final class Eval
{

    /** the preferred minimum distance between states */
    public final float D_MIN = 0.07f;
    /** the preferred mimimum angle between adjoining connections,
		  in radians */
    public final float T_MIN = (float)((20.0*Math.PI)/180);

    /**
     * Give a numerical evaluation of a layout of the states and connections
     * in the given machine. The higher the result, the "better" the layout
     * according to the criteria used. The values returned from each criterion
     * evaluation are transformed in various way and combined, with various
     * coefficients to given an overall evaluation
     * @param m the machine of interest
     * @return the evaluation
     */
    public float evaluation(Logical m) 
    {
		  float c1=1.0f;
		  float c2=13.0f;
		  float c3=3.0f;
		  float c4=0.5f;
		  float c5=2.5f;

		  float overall=0.0f;
		  int n = m.numStates();

		  float Ecross = 1.0f / (1.0f + nCross(m));
		  overall += c1*Ecross;

		  float Estatesep=0.0f;
		  if (n!=0)
				Estatesep = (dStates(m) / (0.5f * n * (n-1))) / (D_MIN*D_MIN);
		  overall += c2*Estatesep;

		  float Eangle = cAngle(m) / (T_MIN*T_MIN);
		  overall += c3*Eangle;

		  float Ecompact = 1.0f / (1.0f + sRect(m));
		  overall += c4*Ecompact;

		  float Econprox = 1.0f / (1.0f + cLength(m));
		  overall += c5*Econprox;

		  return overall;
	 }

    /**
     * Calculate the angle between two lines segments sharing a common point.
     * @param p1 one end of a line
     * @param p2 the shared point
     * @param p3 the end of the other line
     * @return the angle between the two lines
     */
    public float angle(Coord p1,Coord p2,Coord p3) 
    {
		  // the cosine rule
		  float a2 = (float) (Math.pow((p3.x-p1.x),2)+Math.pow(p3.y-p1.y,2));
		  float b2 = (float) (Math.pow((p1.x-p2.x),2)+Math.pow(p1.y-p2.y,2));
		  float c2 = (float) (Math.pow((p3.x-p2.x),2)+Math.pow(p3.y-p2.y,2));
		  float bc = (float) (2*Math.sqrt(b2)*Math.sqrt(c2));
		  float t = (float) ((b2+c2-a2)/bc);
		  float ang;
		  if ((t >= 1.0f) || (t <= -1.0f)) // workaround bug in netscape Sun4
				ang = 0.0f;
		  else
				ang = (float) Math.acos(t);
		  return(ang);
    }

    /**
     * Calculate the total squared angle between adjoining connections,
     * if less than T_MIN, and T_MIN otherwise
     * @param m the machine of interest
     * @return the total angle measure
     */
    public float cAngle(Logical m) 
    {
		  float ca = 0.0f;
		  // count number of connections sharing a common point
		  int p=0;
		  Coord p1,p2,p3;
		  float theta;
		  for (int j=0; j<m.numConnections(); j++) 
		  {
				Connection c1 = m.connectionAt(j);
				for (int i=j+1; i<m.numConnections(); i++) 
				{
					 Connection c2 = m.connectionAt(i);
					 // only do if connections share a common point
					 if ((c1.getFrom() == c2.getFrom()) ||
						  (c1.getTo() == c2.getTo()) ||
						  (c1.getFrom() == c2.getTo()) ||
						  (c1.getTo() == c2.getFrom())) 
					 {
						  p++;
						  if (c1.getFrom() == c2.getFrom()) 
						  {
								p1=c1.getTo().getPos();
								p2=c1.getFrom().getPos();
								p3=c2.getTo().getPos();
						  }
						  else if (c1.getTo() == c2.getTo()) 
						  {
								p1=c1.getFrom().getPos();
								p2=c1.getTo().getPos();
								p3=c2.getFrom().getPos();
						  }
						  else if (c1.getFrom() == c2.getTo()) 
						  {
								p1=c1.getTo().getPos();
								p2=c1.getFrom().getPos();
								p3=c2.getFrom().getPos();
						  }
						  else { // (c1.getTo() == c2.getFrom())
								p1=c1.getFrom().getPos();
								p2=c1.getTo().getPos();	
								p3=c2.getTo().getPos();
						  }
						  theta = angle(p1,p2,p3);
						  if (theta < T_MIN)
								ca += theta*theta;
						  else
								ca += T_MIN*T_MIN;
					 }
				}
		  }
		  if (p==0)
				return 0.0f;
		  return ca/p;
    }

    /**
     * Calculate the number of crossing connections in the given machine
     * @param m the machine of interest
     * @return the number of crossings
     */
    public int nCross(Logical m) 
    {
		  int nc=0;
		  for (int j=0; j<m.numConnections(); j++) 
		  {
				Connection c1 = m.connectionAt(j);
				for (int i=j+1; i<m.numConnections(); i++) 
				{
					 Connection c2 = m.connectionAt(i);
					 // only do if connections don't share a common point
					 if (!((c1.getFrom() == c2.getFrom()) ||
							 (c1.getTo() == c2.getTo()) ||
							 (c1.getFrom() == c2.getTo()) ||
							 (c1.getTo() == c2.getFrom()))) 
					 {
						  // check for intersection
						  if (intersect(c1.getFrom().getPos(),
											 c1.getTo().getPos(),
											 c2.getFrom().getPos(),
											 c2.getTo().getPos()))
								nc++;
					 }
				}
		  }
		  return nc;
    }

    /**
     * Determine whether two line segments intersect.
     * @param l1,l2 the start and end points of one line
     * @param l3,l4 the start and end points of the other line
     * @return true if the line segments intersection, otherwise false
     */
    public boolean intersect(Coord l1,Coord l2,Coord l3,Coord l4) 
    {

		  // convert line segments to y=mx+c and y=nx+d line format
		  float m = (l2.y-l1.y)/(l2.x-l1.x);
		  float c = l1.y - m*l1.x;
		  float n = (l4.y-l3.y)/(l4.x-l3.x);
		  float d = l3.y - n*l3.x;

		  // determine point of intersection of lines
		  float x,y;
		  if (Float.isInfinite(m) && Float.isInfinite(n)) 
		  {
				return(false); // parallel lines - no intersection
		  }
		  else if (Float.isInfinite(m)) { // first line is vertical
				x = l1.x;
				y = n*x + d;
		  }
		  else if (Float.isInfinite(n)) { // second line is vertical
				x = l3.x;
				y = m*x + c;
		  }
		  else 
		  {
				x = (d-c)/(m-n);
				y = m*x + c;
		  }

		  // does point of intersection lie on both line segments?
		  if ((x>=Math.min(l1.x,l2.x)) && (x<=Math.max(l1.x,l2.x)) &&
				(y>=Math.min(l1.y,l2.y)) && (y<=Math.max(l1.y,l2.y)) &&
				(x>=Math.min(l3.x,l4.x)) && (x<=Math.max(l3.x,l4.x)) &&
				(y>=Math.min(l3.y,l4.y)) && (y<=Math.max(l3.y,l4.y)))
				return(true);	
		  return(false);
    }

    /**
     * Calculate the length of all connections in the machine
     * @param m the machine if interest
     * @return the total connection length
     */
    public float cLength(Logical m) 
    {
		  float cl=0.0f;
		  for (int i=0; i<m.numConnections(); i++) 
		  {
				Connection c = m.connectionAt(i);
				cl += c.getFrom().getPos().dist(c.getTo().getPos());
		  }
		  return cl;
    }

    /**
     * Calculate the total squared distance between every pair of states, 
     * if less than D_MIN, and D_MIN otherwise
     * @param m the machine of interest
     * @return the total distance measure
     */
    public float dStates(Logical m) 
    {
		  float ds=0.0f;
		  for (int j=0; j<m.numStates(); j++) 
		  {
				State s1 = m.stateAt(j);
				for (int i=j+1; i<m.numStates(); i++) 
				{
					 State s2 = m.stateAt(i);
					 float d = s1.getPos().dist(s2.getPos());
					 if (d < D_MIN)
						  ds += d*d;
					 else
						  ds += D_MIN*D_MIN;
				}
		  }
		  return ds;
    }

    /**
     * Calculate the area of the smallest enclosing rectangle of all the
     * states in the given machine
     * @param m the machine of interest
     * @return the area of the smallest enclosing rectangle
     */
    public float sRect(Logical m) 
    {
		  if (m.numStates() > 0) 
		  {
				float left,right,top,bottom;
				State s = m.stateAt(0);
				Coord p = s.getPos();
				left = p.x; right = p.x;
				top = p.y; bottom = p.y;
				for (int i=1; i<m.numStates(); i++) 
				{
					 s = m.stateAt(i);
					 p = s.getPos();
					 if (p.x < left) left = p.x;
					 if (p.x > right) right = p.x;
					 if (p.y < top) top = p.y;
					 if (p.y > bottom) bottom = p.y;
				}
				return (right-left)*(bottom-top);
		  }
		  return 0.0f;
    }
}
