/* 
 * The Finite State Machine Explorer
 * Copyright (c) 1995-1996
 * Matt Chapman (csuoq@csv.warwick.ac.uk)
 * Dept. Computer Science
 * University of Warwick, Coventry, UK.
 */

package Machine;

import java.util.*;

/**
 * Attempts to layout the states and connections of a machine in a "nice" way
 * @version 19 Feb 1996
 * @author Matt Chapman
 */
public final class Layout 
{
    /**
     * Generate a layout for the given machine.
     * It works by fixing the start state and then taking another state,
     * placing it in the best surrounding position, according to the
     * evaluation function used. This is then repeated for all of the states.
     * Several different orderings of the states are tried, the one with
     * the highest overall evaluation being chosen. Ideally all possible
     * orderings should be testest, but this would be too slow, with
     * factorial complexity.
     * @param m the machine to be laid out
     */
    public void layout(Logical m) 
    {
		  int NTRIES=4;
		  Eval eval = new Eval();
		  float tsize = 0.07f; // D_MIN from Eval

		  int ns = m.numStates();
		  float bx[] = new float[ns];
		  float by[] = new float[ns];
		  float btotal = 0;

		  for (int n=0; n<NTRIES; n++)
		  {

				// create a set of all the states in the machine
				Set states = new Set();
				for (int i=0; i<m.numStates(); i++) 
				{
					 State s = m.stateAt(i);
					 states.addElement(s);
				}
				Set placed = new Set();

				// put start state to left hand side
				State start = m.getStart();
				Coord p = new Coord(0.1f,0.1f);
				Coord startpos = new Coord(p);
				start.setPos(p);
				states.removeElement(start);

				Logical nm = new Logical();
				nm.addGiven(start);
				nm.makeStart(start);
				placed.addElement(start);

				int xoff[] = {1,1,0,-1,-1,-1,0,1};
				int yoff[] = {0,1,1,1,0,-1,-1,-1};
				while (states.size()>0)
				{
					 State s = rndState(states);
					 states.removeElement(s);
					 // add state to new machine
					 nm.addGiven(s);
					 placed.addElement(s);
					 // add all connections involving states present
					 for (int j=0; j<m.numConnections(); j++) 
					 {
						  Connection c = m.connectionAt(j);
						  State from = c.getFrom();
						  State to = c.getTo();
						  if (placed.contains(from) && placed.contains(to))
								nm.addConnection(from,to,c.getTrans());
					 }
					 float curr,best=0;
					 Coord bestpos = new Coord();
					 // try placing state in each of 8 positions
					 for (int i=0; i<8; i++)
					 {
						  float xd = tsize * xoff[i];
						  float yd = tsize * yoff[i];
						  Coord next = new Coord(p.x+xd,p.y+yd);
						  if ((next.x >= startpos.x) && (next.x < 1.0f)
								&& (next.y < 0.6f) && (next.x > 0.0f)
								&& (next.y > 0.0f))
						  {
								// position state
								s.setPos(next);
								// evaluate current layout
								curr = eval.evaluation(nm);
								if (curr > best)
								{
									 best = curr;
									 bestpos = next;
								}
						  }
					 }
					 s.setPos(bestpos);
					 p=bestpos;
				}
				float ctotal = eval.evaluation(nm);
				if (ctotal > btotal)
				{
					 btotal = ctotal;
					 // store positions
					 for (int k=0; k<ns; k++) 
					 {
						  State s = m.stateAt(k);
						  Coord c = s.getPos();
						  bx[k]=c.x;
						  by[k]=c.y;
					 }
				}
		  }

		  // restore best layout
		  for (int k=0; k<ns; k++) 
		  {
				State s = m.stateAt(k);
				Coord c = new Coord(bx[k],by[k]);
				s.setPos(c);
		  }
    }

    // return random state from a set of states
    private State rndState(Set s)
    {
		  int ch = (int)Math.ceil(0.5 + Math.random()*(s.size()-1)) - 1;
		  return (State)s.elementAt(ch);
    }

    // improve layout by successive movement of states
    public void improve(Logical m) 
    {
		  Eval eval = new Eval();
		  for (float tsize = 4.0f; tsize >= 0.5f; tsize /= 2.0f) 
		  {
				float st = eval.evaluation(m);
				boolean changed = true;
				while (changed) 
				{
					 float start = eval.evaluation(m);
					 float ebest = start;
					 float e;
					 for (int i=0; i<m.numStates(); i++) 
					 {
						  State s = m.stateAt(i);
						  Coord p = s.getPos();
						  float dia = tsize*s.radius;
						  Coord best = new Coord(p.x,p.y);
						  Coord left = new Coord(p.x-dia,p.y);
						  Coord right = new Coord(p.x+dia,p.y);
						  Coord up = new Coord(p.x,p.y-dia);
						  Coord down = new Coord(p.x,p.y+dia);
						  s.setPos(left); e = eval.evaluation(m);
						  if (e>ebest) { best = left;	ebest = e; }
						  s.setPos(right); e = eval.evaluation(m);
						  if (e>ebest) { best = right; ebest = e; }
						  s.setPos(up); e = eval.evaluation(m);
						  if (e>ebest) { best = up; ebest = e; }
						  s.setPos(down); e = eval.evaluation(m);
						  if (e>ebest) { best = down; ebest = e; }
						  s.setPos(best);
					 }
					 if (start==ebest) 
					 {
						  changed = false;
					 }
				}
		  }
    }
}
