/* 
 * The Finite State Machine Explorer
 * Copyright (c) 1995-1996
 * Matt Chapman (csuoq@csv.warwick.ac.uk)
 * Dept. Computer Science
 * University of Warwick, Coventry, UK.
 */

package Machine;

import java.util.Vector;

/**
 * Represents the logical structure of a Finite State Machine.
 * @version    21 Jan 1996
 * @author     Matt Chapman
 */
public final class Logical 
{

    /*
     * The logical structure is stored in a Vector of the States and a Vector
     * of Connections.
     */
    private Vector stateV;
    private Vector connectV;

    private State startState = null;

    /**
     * Creates a new, empty logical machine.
     */
    public Logical() 
    {
		  stateV = new Vector();
		  connectV = new Vector();
    }

    /**
     * Adds a new state to the current logical structure.
     * @return the state added
     */
    public State addState() 
    {
		  State s = new State();
		  stateV.addElement(s);
		  return s;
    }

    public void addGiven(State s) 
    {
		  stateV.addElement(s);
    }

    /**
     * Removes a state from the current logical strucure, and also removes all
     * connections involving that state.
     * @param s the state to be removed
     */
    public void removeState(State s) 
    {
		  // remove connections involving state s

		  // clone the connection vector so that removing connections doesn't
		  // affect the traversal of connections
		  Vector cl = (Vector)connectV.clone();
		  for (int i=0; i<cl.size(); i++) 
		  {
				Connection c = (Connection)cl.elementAt(i);
				if ((c.getFrom() == s) || (c.getTo() == s))
					 removeConnection(c);
		  }

		  // remove start designation if s is the start state
		  if (s == startState)
				startState = null;

		  // remove s itself
		  stateV.removeElement(s);
    }
 
    /**
     * Add a connection between two states, for a particular transition
     * character.
     * @param from the start state for the connection
     * @param to the destination state for the connection
     * @param t the transition character for the connection
     */
    public void addConnection(State from, State to, char t) 
    {
		  // see if a connection between given two states already exists
		  for (int i=0; i<numConnections(); i++) 
		  {
				Connection c = connectionAt(i);
				if ((c.getFrom() == from) && (c.getTo() == to)) 
				{
					 String old = c.getTrans();
					 if (old.indexOf(t) == -1)
						  c.setTrans(c.getTrans()+t);
					 return;
				}
		  }

		  Connection c = new Connection(from,to,t+"");
		  connectV.addElement(c);
    }

    public void addConnection(State from, State to, String str)
    {
		  for (int i=0; i<str.length(); i++)
		  {
				addConnection(from,to,str.charAt(i));
		  }
    }

    /**
     * Remove a connection from the current logical structure.
     * @param c the connection to be removed
     */
    public void removeConnection(Connection c) 
    {
		  connectV.removeElement(c);
    }

    /**
     * Returns the destination state given a current state and a transiton
     * character. If there is more than one such state, any one is returned.
     * @param s the current state
     * @param s the transition character
     * @return a destination state, or null if none exist
     */
    public State transition(State s, char t) 
    {
		  // search all connections for required transition
		  for (int i=0; i<connectV.size(); i++) 
		  {
				Connection c = connectionAt(i);
				if ((c.getFrom() == s) && (c.getTrans().indexOf(t) > -1))
					 return c.getTo();				      
		  }
		  return null; // no relevant transition
    }

    /**
     * Return the set of destination states given a current state and a
     * transiton character.
     * @param s the current state
     * @param s the transition character
     * @return the set (possible empty) of destination states
     */
    public Set transitionSet(State s, char t) 
    {
		  Set dest = new Set();
		  // search all connections for required transition
		  for (int i=0; i<connectV.size(); i++) 
		  {
				Connection c = connectionAt(i);
				if ((c.getFrom() == s) && (c.getTrans().indexOf(t) > -1))
					 dest.addElement(c.getTo());
		  }
		  return dest;
    }

    /**
     * Designate a state to be the start state of the machine.
     * If a different state was previously defined to be the start state,
     * it will lose this designation.
     * @param s the state to make the start state
     */
    public void makeStart(State s) 
    {
		  startState=s;
    }

    /**
     * Returns the start state of the machine.
     * @return the start state
     */
    public State getStart() 
    {
		  return startState;
    }

    /**
     * Returns the total number of states in the machine.
     * @return the number of states
     */
    public int numStates() 
    {
		  return stateV.size();
    }

    /**
     * Returns the nth state of the machine.
     * @param n the number of the state required
     * @return the state which is labelled by n
     */
    public State stateAt(int n) 
    {
		  return (State)stateV.elementAt(n);
    }

    /**
     * Returns the total number of connections in the machine.
     * @return the number of connections
     */
    public int numConnections() 
    {
		  return connectV.size();
    }

    /**
     * Returns the nth connection of the machine.
     * @param n the number of the connection required
     * @return the connection which is labelled by n
     */
    public Connection connectionAt(int n) 
    {
		  return (Connection)connectV.elementAt(n);
    }

}



