/* 
 * The Finite State Machine Explorer
 * Copyright (c) 1995-1996
 * Matt Chapman (csuoq@csv.warwick.ac.uk)
 * Dept. Computer Science
 * University of Warwick, Coventry, UK.
 */

package Machine;

import java.io.*;
import java.util.Vector;

/**
 * Input and Output operations for machines
 * @version    1 March 1996
 * @author     Matt Chapman
 */
public final class MachineIO
{
    // constants for lexing phase
    private final int NONE=0;
    private final int STATE_ID=1;
    private final int STATE_X=2;
    private final int STATE_Y=3;
    private final int STATE_ACC=4;
    private final int CONNECT_FROM=5;
    private final int CONNECT_TO=6;
    private final int CONNECT_STR=7;
    private final int START=8;

    private static PrintStream out;
    private static StreamTokenizer in;

    /**
     * Save a machine to a file
     * @param m the machine to be saved
     * @param filename the filename to save the machine as
     */
    public void save(Logical m, String filename)
    {
		  // open file for output
        try { 
            out = new PrintStream(new FileOutputStream(filename));
        }
        catch (Exception e) {
            System.out.println("Could not save to \"" + filename + "\".");
            System.out.println("Exception: " + e);
            return;
        }

		  out.println("# A saved machine"); 

	// save state details
		  out.println("# state details");
		  for (int i=0; i<m.numStates(); i++) 
		  {
				State s = m.stateAt(i);
				Coord p = s.getPos();
				out.println("State "+s.getID()+" "+p.x+" "+p.y+" "
								+s.isAccepting());
		  }

		  // save connection details
		  out.println("# connection details");
		  for (int i=0; i<m.numConnections(); i++) 
		  {
				Connection c = m.connectionAt(i);
				c.calculatePos();
				out.println("Connection "+c.getFrom().getID()+" "
								+c.getTo().getID()+" C"+c.getTrans());
		  }
		  out.println("# start state, if defined");
		  if (m.getStart() != null)
				out.println("Start "+m.getStart().getID());
		  out.println("# end of saved machine");

		  try
		  {
				out.close();
		  }
		  catch (Exception e)
		  {
				return;
		  }
    }

    /**
     * Load a previously saved machine
     * @param the filename to load the machine from
     * @return the machine contructed from the loaded file
     */
    public Logical load(String filename)
    {
		  Logical m = new Logical();

		  // open file for reading
        try { 
            in = new StreamTokenizer(new FileInputStream(filename));
        }
        catch (Exception e) {
            System.out.println("Could not save to \"" + filename + "\".");
            System.out.println("Exception: " + e);
            return null;
        }

		  in.commentChar('#');

		  int token;
		  int mode=NONE;

		  int state_id = 0;
		  float state_x = 0.0f;
		  float state_y = 0.0f;
		  boolean accept = false;
		  int con_from = 0,con_to = 0;
		  String con_str = "";
		  int start_id = 0;

		  Vector stateV = new Vector();

		  try
		  {
				token = in.nextToken();
		  }
		  catch (Exception e)
		  {
				System.out.println("Error reading token");
				return null;
		  }
		  while (token != StreamTokenizer.TT_EOF)
		  {
				if (token == StreamTokenizer.TT_WORD)
				{
					 switch(mode)
					 {
					 case NONE:
					 {
						  if (in.sval.equals("State"))
								mode = STATE_ID;
						  else if (in.sval.equals("Connection"))
								mode = CONNECT_FROM;
						  else if (in.sval.equals("Start"))
								mode = START;
						  break;
					 }
					 case STATE_ACC:
					 {
						  if (in.sval.equals("true"))
								accept = true;
						  else
								accept = false;

						  // add state to machine
						  State s = m.addState();
						  Coord c = new Coord(state_x,state_y);
						  s.setPos(c);
						  if (accept)
								s.makeAccepting();

						  // increase vector size, if not big enough
						  if (stateV.size() < state_id+1)
								stateV.setSize(state_id*2);
						  // add state to vector
						  stateV.setElementAt(s,state_id);
						  mode = NONE;
						  break;
					 }	
					 case CONNECT_STR:
					 {
						  con_str = in.sval.substring(1);

						  // add connection to machine
						  m.addConnection((State)stateV.elementAt(con_from),
												(State)stateV.elementAt(con_to),
												con_str);
						  mode = NONE;
						  break;
					 }
					 }
				}
				if (token == StreamTokenizer.TT_NUMBER)
				{
					 switch(mode)
					 {
					 case STATE_ID:
					 {
						  state_id = (int)in.nval;
						  mode++;
						  break;
					 }
					 case STATE_X:
					 {
						  state_x = (float)in.nval;
						  mode++;
						  break;
					 }
					 case STATE_Y:
					 {
						  state_y = (float)in.nval;
						  mode++;
						  break;
					 }
					 case CONNECT_FROM:
					 {
						  con_from = (int)in.nval;
						  mode++;
						  break;
					 }
					 case CONNECT_TO:
					 {
						  con_to = (int)in.nval;
						  mode++;
						  break;
					 }
					 case START:
					 {
						  start_id = (int)in.nval;
						  mode = NONE;

						  // designate start state in machine
						  m.makeStart((State)stateV.elementAt(start_id));

						  break; 
					 }
					 }
				}
				try
				{
					 token = in.nextToken();
				}
				catch (Exception e)
				{
					 System.out.println("Error reading token");
					 return null;
				}
		  }
		  return m;
    }

}
