/* 
 * The Finite State Machine Explorer
 * Copyright (c) 1995-1996
 * Matt Chapman (csuoq@csv.warwick.ac.uk)
 * Dept. Computer Science
 * University of Warwick, Coventry, UK.
 */

package Machine;

import java.awt.*;
import java.awt.event.*;
import java.util.*;
import java.applet.Applet;
import java.net.*;

/**
 * The main widget for the Finite State Machine Explorer
 * @version    24 Nov 1996
 * @author     Matt Chapman
 */
public final class MainWidget extends Canvas 
{

    // constants
    final int ADD=1, MOVE=2, CONNECT=3, DELETE=4, START=5, TOGGLE=6;
    final int wid=570, hei=342;
    final int twid=1140, thei=600;

    // initialise other classes
    Logical machine = new Logical();
    World screen = new World();
    Eval eval = new Eval();
    Simulate sim;
    Layout lay = new Layout();
    Minimize min = new Minimize();
    Convert conv = new Convert();
    Postscript postscript = new Postscript();
    MachineIO mcIO = new MachineIO();

    // information about current situation of machine design
    int mode=ADD;
    int to_state;
    int selected_state;
    Coord move_orig;

    State moving_state;
    State connect_state;
    State connect_to;

    // interface components
    Checkbox cbox;
    Button simbutton;
    TextField inputfield;
    Choice simspeed;
    WPanner pannerw;
    ZoomSlide zoomw;
    Frame frame = null;
    Image buttons;
    Label msg_label;

    Dialog convd;
    ConvThread convThread;

    // flags
    boolean making_connection = false;
    boolean waiting_for_key = false;
    boolean waiting_for_transition = false;
    boolean during_move = false;
    boolean empty_drag = false;
    boolean selecting_start = false;
    boolean make_accepting = false;

    // for double buffering to prevent flicker
    Image offScreenImage;
    Graphics offScreen;

    public MainWidget() 
    {
	reshape(0, 0, wid, hei);
    }

    public void msg(String txt) 
    {
	msg_label.setText(txt);
    }

    // define the label widget to be used for messages
    public void setMsg(Label msg) 
    {
	msg_label = msg;
    }

    // construct a machine for test purposes
    private void constructTest() 
    {
	machine = new Logical();
	State s1 = machine.addState();
	State s2 = machine.addState();
	State s3 = machine.addState();
	State s4 = machine.addState();
	State s5 = machine.addState();
	State s6 = machine.addState();
	machine.addConnection(s1,s2,'a');
	machine.addConnection(s1,s3,'b');
	machine.addConnection(s2,s2,'a');
	machine.addConnection(s2,s4,'b');
	machine.addConnection(s3,s4,'a');
	machine.addConnection(s3,s5,'b');
	machine.addConnection(s4,s2,'a');
	machine.addConnection(s4,s6,'b');
	machine.addConnection(s5,s5,'a');
	machine.addConnection(s5,s4,'b');
	machine.addConnection(s6,s6,'a');
	machine.addConnection(s6,s6,'b');
	s2.makeAccepting();
	s5.makeAccepting();
	s6.makeAccepting();
	machine.makeStart(s1);
    }

    public boolean keyDown(Event evt, int key) 
    {
	char k = (char)key;

	if (waiting_for_transition) 
	{
	    waiting_for_transition = false;
	    if ((k>='a' && k<='z') || (k>='A' && k<='Z')
		|| (k>='0' && k<='9')) 
	    {
		machine.addConnection(connect_state,connect_to,k);
		msg("States connected.");
	    }
	    else 
	    {
		msg("Invalid character. Connection cancelled.");
	    }
	    screen.hilight = null;
	    repaint();
	    return false;
	}

        if (k=='a') 
        {
	    changeMode(ADD);
	}
	else if (k=='m') 
	{
	    changeMode(MOVE);
	}
	else if (k=='c') 
	{
	    changeMode(CONNECT);
	}
	else if (k=='t') 
	{
	    changeMode(TOGGLE);
	}
	else if (k=='d') 
	{
	    changeMode(DELETE);
	}
	else if (k=='s') 
	{
	    changeMode(START);
	}
	else if (k=='e')
	{ // to test evaluation function
	    float e = eval.evaluation(machine);
	    System.out.println("E="+e);
	}
	else if (k=='l')
	{ // construct a machine, for testing
	    constructTest();
	}
	else if (k=='i')
	{ // call improve function
	    lay.improve(machine);
	}
	repaint();
	return false;
    }

    /**
     * Call the layout class for the current machine
     */
    public void doLayout()
    {
	if (machine.numStates() == 0)
	{
	    msg("No machine to layout");
	}
	else
	{
	    if (machine.getStart() == null)
	    {
		msg("The machine must have a start state defined");
	    }
	    else
	    {
		setBusy();
		lay.layout(machine);
		msg("Layout complete");
		clearBusy();
	    }
	}
	repaint();
    }

    /**
     * Save the machine to a file
     * @param filename the file the machine is to be saved in
     */
    public void saveMC(String filename)
    {
	repaint();
	mcIO.save(machine,filename);
	msg("Machine saved");
    }

    /**
     * Load a machine from a file
     * @param filename the file to be loaded as a machine
     */
    public void loadMC(String filename)
    {
	repaint();
	Logical m = mcIO.load(filename);
	if (m != null)
	{
	    machine = m;
	    msg("Machine loaded");
	}
	else
	    msg("Error during load");
	repaint();
    }

    /**
     * Output a PostScript version of the machine
     * @param filename where to save the PostScript
     * @param title an optional title for the output
     * @param portrait the orientation of the output, true for portrait,
     * false for landscape
     */
    public void genPS(String filename, String title, boolean portrait)
    {
	repaint();
	postscript.PSpaint(machine,filename,title,portrait);
	msg("PostScript output generated");
    }

    /**
     * Clear the current machine
     */
    public void startNew()
    {
	Logical nm = new Logical();
	machine = nm;
	repaint();
    }

    /**
     * Determine whether the current machine is deterministic
     */
    public void isDet()
    {
	if (machine.numStates() == 0)
	{
	    msg("No current machine");
	}
	else
	{
	    if (conv.isDeterministic(machine))
	    {
		msg("The machine is deterministic");
	    }
	    else
	    {
		msg("The machine is non-deterministic");
	    }
	}
	repaint();
    }

    /**
     * Remove inaccessible states
     */
    public void remInacc()
    {
	if (machine.numStates() == 0)
	{
	    msg("No current machine");
	}
	else
	{
	    if (machine.getStart() == null)
	    {
		msg("The machine must have a start state defined");
	    }
	    else
	    {
		// remove inaccessible states		
		min.removeInacc(machine);		
		msg("All inaccessible states have been removed");
	    }
	}
	repaint();
    }

    /**
     * Call the minimization class
     */
    public void minimize()
    {
	if (machine.numStates() == 0)
	{
	    msg("No machine to minimize");
	}
	else
	{
	    if (machine.getStart() == null)
	    {
		msg("The machine must have a start state defined");
	    }
	    else
	    {
		if (conv.isDeterministic(machine))
		{
		    setBusy();
		    // remove inaccessible states
		    min.removeInacc(machine);
		    // make machine complete
		    min.complete(machine);
		    // merge indistinguishable states
		    min.minimum(machine);
		    lay.layout(machine);
		    msg("The machine is now minimal");
		    clearBusy();
		}
		else
		{
		    msg("The machine must first be made deterministic");
		}
	    }
	}
	repaint();
    }

    /**
     * Convert from NFA to DFA
     */
    public void nfa()    
    {
	if (machine.numStates() == 0)
	{
	    msg("No machine to convert");
	}
	else
	{
	    if (machine.getStart() == null)
	    {
		msg("The machine must have a start state defined");
	    }
	    else
	    {
		setBusy();
					 
		Component c = getParent();
		while (c!=null && !(c instanceof Frame)) {
		    c = c.getParent();
		}
		if (c!=null) {
		    convd = new Dialog((Frame)c,true);
		    Panel p = new Panel();
		    convd.add(p);
		    p.add(new Label("Conversion to deterministic in progress..."));
		    Button b = new Button("Cancel");
		    b.addActionListener(new ActionListener() {
			    public void actionPerformed(ActionEvent e) {
				if (convThread!=null) {
				    convThread.stop();
				    convThread=null;
				}
				if (convd!=null) {
				    convd.setVisible(false);
				    convd.dispose();
				    convd=null;
				}
				clearBusy();
			    }
			});
		    p.add(b);
		    convd.pack();

		    convThread = new ConvThread(this);
		    convThread.start();

		    convd.show();
		}
	    }
	}
	repaint();
    }

    public void changeMode(int newmode) 
    {
	if (newmode>0 && newmode<=6 && newmode!=mode) 
	{
	    mode=newmode;
	    setCursor(Frame.DEFAULT_CURSOR);
	    making_connection = false;
	    screen.hilight = null;
	    waiting_for_transition = false;
	    if (mode==ADD) 
	    {
		msg("Click to add states...");
	    }
	    else if (mode==MOVE) 
	    {
		msg("Move states or connections by dragging them...");
	    }
	    else if (mode==CONNECT) 
	    {
		msg("Select a state to initiate connection...");
	    }
	    else if (mode==TOGGLE) 
	    {
		msg("Select states to toggle their acceptance...");
	    }
	    else if (mode==DELETE) 
	    {
		msg("Select states or connections to delete them...");
		setCursor(Frame.CROSSHAIR_CURSOR);
	    }
	    else if (mode==START) 
	    {
		msg("Select the state to be the starting state...");
	    }
	}
    }

    public boolean lostFocus(Event e, Object obj) 
    {
	return false;
    }

    public boolean mouseEnter(Event e, int x, int y) 
    {
	requestFocus();
	return false;
    }

    public boolean mouseDrag(Event e, int x, int y) 
    {
	if (mode==MOVE) 
	{
	    Point p = new Point(x,y);
	    Coord c = screen.reversePipeline(p);
	    if (during_move) 
	    {
		moving_state.setPos(c);
	    }
	    repaint();
	    return true;
	}
	return false;
    }

    public boolean mouseDown(Event e, int x, int y) 
    {
	int b = inbuttons(x,y);
	if (b>0) 
	{
	    changeMode(b);
	    repaint();
	    return true;
	}

	if (mode==MOVE)
	{ // start move operation
	    Point p = new Point(x,y);
	    Coord c = screen.reversePipeline(p);
	    State s = screen.getState(machine,p);
	    if (s != null) 
	    {
		during_move = true;
		move_orig = s.getPos(); // store original position
		moving_state = s;
		setCursor(Frame.MOVE_CURSOR);
	    }
	    return true;
	}
	return false;
    }

    public boolean mouseUp(Event e, int x, int y) 
    {
	int b = inbuttons(x,y);
	if (b>0 && mode!=MOVE) 
	{
	    return true;
	}

	if (mode==ADD) 
	{
	    Point p = new Point(x,y);
	    Coord c = screen.reversePipeline(p);
	    
	    if (!screen.roomFor(machine,p)) 
	    {
		msg("There is no room at that position.");
	    }
	    else 
	    {
		// add state
		State s = machine.addState();
		s.setPos(c);
		//paint(getGraphics());
	    }
	}
	else if (mode==MOVE) 
	{
	    during_move = false;
	    setCursor(Frame.DEFAULT_CURSOR);
	    Point p = new Point(x,y);
	    // check for space at the specified location
	    if (!screen.roomFor(machine,p,moving_state)) 
	    {
		// move back to orginal position
		moving_state.setPos(move_orig);
		msg("There is no room at that position.");
	    }
	}
	else if (mode==CONNECT) 
	{
	    Point p = new Point(x,y);
	    State s = screen.getState(machine,p);
	    if (s != null) 
	    {
		if (!making_connection) 
		{
		    msg("Select state to connect this one to...");
		    setCursor(Frame.HAND_CURSOR);
		    making_connection = true;
		    connect_state = s;
		    Set hset = new Set();
		    hset.addElement(connect_state);
		    screen.hilight = hset;
		}
		else 
		{
		    setCursor(Frame.DEFAULT_CURSOR);
		    making_connection = false;
		    waiting_for_transition = true;
		    connect_to = s;
		    msg("Press a key for that transition...");
		    screen.hilight.addElement(connect_to);
		}
	    }
	}
	else if (mode==TOGGLE) 
	{
	    Point p = new Point(x,y);
	    State s = screen.getState(machine,p);
	    if (s != null) 
	    {
		if (s.isAccepting())
		    s.makeRejecting();
		else
		    s.makeAccepting();
	    }
	}
	else if (mode==DELETE) 
	{
	    Point p = new Point(x,y);
	    State s = screen.getState(machine,p);
	    if (s != null) 
	    {
		machine.removeState(s);
	    }
	    else 
	    {
		Connection c = screen.getConnection(machine,p);
		if (c != null) 
		{
		    machine.removeConnection(c);
		}
	    }
	}
	else if (mode==START) 
	{
	    Point p = new Point(x,y);
	    State s = screen.getState(machine,p);
	    if (s != null) 
	    {
		machine.makeStart(s);
	    } 
	}
	repaint();
	return true;
    }

    public void markStart(Checkbox c) 
    {
	cbox = c;
	msg("Select the starting state...");
	selecting_start = true;
	// cancel any other modes
	make_accepting = false;
	making_connection = false;
	waiting_for_transition = false;
    }

    public void makeAccepting(Checkbox c) 
    {
	cbox = c;
	msg("Select accepting states...");
	make_accepting = true;
	// cancel any other modes
	selecting_start = false;
	making_connection = false;
	waiting_for_transition = false;
    }

    public void togglelabels() 
    {
	repaint();
    }

    // change the current zoom
    public void setZoom(float mag, boolean show) 
    {
	screen.setScaling(mag);
	pannerw.msetSize((int)(wid*mag),(int)(hei*mag));
	pannerw.paint(pannerw.getGraphics());
	setWindow(pannerw.getXValue(),pannerw.getYValue(),false);

	repaint();
    }

    // change the current viewport
    public void setWindow(int xwin, int ywin, boolean show) 
    {
	screen.setWindow(xwin,ywin);

	repaint();
    }

    // call the simulation class
    public void simulate() 
    {
	sim = new Simulate(msg_label);
	boolean validI = sim.validInput(inputfield.getText());
	if (validI) 
	{
	    boolean validM = sim.validMachine(machine);
	    if (validM) 
	    {
		repaint();
		sim.runSim(inputfield.getText(),this.getGraphics(),this,
			   screen,machine,simbutton,simspeed);
	    }
	}
    }

    public void stopsim() 
    {
	sim.stop();
    }

    public void update(Graphics g) 
    {
	paint(g);
    }

    // check for buttons in the main window
    public int inbuttons(int x,int y) 
    {
	if (y>0 && y<buttons.getHeight(this) && x>wid-buttons.getWidth(this)
	    && x<wid)
	    return 1 + y/((buttons.getHeight(this)-1)/6);
	else
	    return 0;
    }

    // repaint the widget
    public synchronized void paint (Graphics g) 
    {
	if (offScreenImage == null) 
	{
	    offScreenImage = createImage (wid,hei);
	    offScreen = offScreenImage.getGraphics();
	    screen.buttons = buttons;
	}
	screen.paintMachine(machine,offScreen,wid,hei,mode);
	g.drawImage(offScreenImage, 0, 0, this);
    }

    // change cursor to busy symbol
    public void setBusy()
    {
	setCursor(Frame.WAIT_CURSOR);
	paint(getGraphics());
    }

    // change cursor back to default
    public void clearBusy()
    {
	setCursor(Frame.DEFAULT_CURSOR);
    }

    // change to given cursor
    public void setCursor(int c) 
    {
	if (frame == null) 
	{
	    Object f = getParent();
	    while (! (f instanceof Frame))
		f = ((Component) f).getParent();
	    frame = (Frame)f;
	}
	frame.setCursor(c);
    }

}


class ConvThread extends Thread {
    MainWidget mw;
    public ConvThread(MainWidget mw) {
	this.mw=mw;
    }
    public void run() {
	// convert to deterministic
	Logical nm = mw.conv.NFA2DFA(mw.machine);
	// remove inaccessible states
	mw.min.removeInacc(nm);
	mw.machine = nm;
	// layout new machine
	mw.lay.layout(mw.machine);
	mw.msg("Conversion to deterministic complete");
	mw.clearBusy();
	if (mw.convd!=null) {
	    mw.convd.setVisible(false);
	    mw.convd.dispose();
	    mw.convd=null;
	}
	mw.repaint();
    }
}
