/* 
 * The Finite State Machine Explorer
 * Copyright (c) 1995-1996
 * Matt Chapman (csuoq@csv.warwick.ac.uk)
 * Dept. Computer Science
 * University of Warwick, Coventry, UK.
 */

package Machine;

/**
 * Routines to minimize a Finite State Machine, including the removal of
 * inaccessible states.
 * @version   22 Feb 1996
 * @author    Matt Chapman
 */
public final class Minimize 
{

    // a set of states - all accessible states are removed until only
    // inaccessible states are left  
    private Set acc = new Set();

    /**
     * Removes all inaccessible states from the given machine.
     * @param m the machine of interest
     */
    public void removeInacc(Logical m) 
    {
		  // create a set of all the states in the machine
		  for (int i=0; i<m.numStates(); i++) 
		  {
				State s = m.stateAt(i);
				acc.addElement(s);
		  }
		  // traverse machine, removing all accessible states from set
		  rec(m,m.getStart());

		  // anything left in set must be inaccessible, so remove from machine
		  for (int i=0; i<acc.size(); i++) 
		  {
				State r = (State)acc.elementAt(i);
				m.removeState(r);
		  }
    }

    /**
     * Remove a state from the current set of states, and recursively
     * remove all states directly connected to it.
     * @param m the machine of interest
     * @param s the state to be removed - it is definitely accessible
     */
    private void rec(Logical m,State s) 
    {
		  if (acc.contains(s)) 
		  {
				acc.removeElement(s);
				Set st = allReach(m,s);
				for (int i=0; i<st.size(); i++) 
				{
					 State r = (State)st.elementAt(i);
					 rec(m,r);
				}
		  }
    }

    /** 
     * Returns a set of all the states that are reachable in exactly one
     * transition from the given state.
     * @param m the machine of interest
     * @param s the state of interest
     * @return the set of reachable states
     */
    private Set allReach(Logical m, State s) 
    {
		  Set reach = new Set();
		  for (int i=0; i<m.numConnections(); i++) 
		  {
				Connection c = m.connectionAt(i);
				if (c.getFrom() == s)
					 reach.addElement(c.getTo());
		  }
		  return reach;
    }

    /**
     * Complete the given machine, by adding a "dead" state, and connects
     * to this state whenever a transition is not already defined
     * @param m the machine to be made complete
     */
    public void complete(Logical m)
    {
		  Convert conv = new Convert();
		  Set ta = conv.transitionAlph(m);
		  State dead = null;
		  for (int i=0; i<m.numStates(); i++) 
		  {
				State s = m.stateAt(i);
				for (int j=0; j<ta.size(); j++)
				{
					 Character ct = (Character)ta.elementAt(j);
					 char t = ct.charValue();
					 if (m.transition(s,t) == null)
					 {
						  // create dead state if necessary
						  if (dead == null)
						  {
								dead = m.addState();
								for (int k=0; k<ta.size(); k++)
								{
									 Character cu = (Character)ta.elementAt(k);
									 char u = cu.charValue();
									 m.addConnection(dead,dead,u);
								}
						  }
						  m.addConnection(s,dead,t);
					 }
				}
		  }
    }

    /**
     * Minimize the given machine, by merging indistinguishable states.
     * @param m the machine to be minimized
     */
    public void minimum(Logical m)
    {
		  // form 0-indistinguishability sets from set of terminal states and
		  // its compliment
		  Set T = new Set();
		  Set Tc = new Set();
		  for (int i=0; i<m.numStates(); i++) 
		  {
				State s = m.stateAt(i);
				if (s.isAccepting())
					 T.addElement(s);
				else
					 Tc.addElement(s);
		  }
		  Convert conv = new Convert();
		  Set ta = conv.transitionAlph(m);

		  Set prev = new Set();
		  Set next = new Set();
		  prev.addElement(T);
		  prev.addElement(Tc);

		  // store start state
		  State start = m.getStart();

		  // generate successive equivalence classes until no extra elements are
		  // added - the final equivalence class is then reached. This forms the
		  // states of the minimal machine
		  boolean done = false;
		  while (!done)
		  {
				next = nextIndis(prev,ta,m);

				if (next.size() == prev.size())
					 done = true;
				prev = next;
		  }

		  // replace indistinguishable states with a single, merged state
		  for (int i=0; i<next.size(); i++)
		  {
				// each element of the set of states of the minimal machine
				Set each = (Set)next.elementAt(i);
				if (each.size() > 1)
				{
					 State merged = m.addState();
					 for (int j=0; j<each.size(); j++)
					 {
						  State s = (State)each.elementAt(j);
						  // if any element of set is accepting then they all are
						  // - so only check first element
						  if ((j==0) && s.isAccepting())
								merged.makeAccepting();
						  // preserve start state
						  if (s == start)
								m.makeStart(merged);
						  for (int k=0; k<m.numConnections(); k++)
						  {
								Connection c = m.connectionAt(k);
								if (c.getTo() == s)
								{
									 m.addConnection(c.getFrom(),merged,c.getTrans());
								}
								if (c.getFrom() == s)
								{
									 m.addConnection(merged,c.getTo(),c.getTrans());
								}
						  }
						  m.removeState(s);
					 }
				}
		  }
    }

    /**
     * Given an equivalence class, generate the subsequent one.
     * @param indis0 the equivalence class to start with
     * @param ta a set of the transition alphabet of the machine
     * @param m the machine for which the equivalence class is being generated
     * @return the subsequent equivalence class
     */
    private Set nextIndis(Set indis0,Set ta,Logical m)
    {
		  // for the subsequent equivalence class
		  Set indis1 = new Set();

		  for (int i=0; i<indis0.size(); i++)
		  {
				Set each = (Set)indis0.elementAt(i);
				if (each.size() > 1)
				{
					 Set nEq = new Set();
					 for (int j=0; j<each.size(); j++)
					 {
						  State s = (State)each.elementAt(j);
						  Set t = new Set();
						  t.addElement(s);
						  nEq.addElement(t);
					 }
					 for (int j=0; j<each.size()-1; j++)
					 {
						  State p = (State)each.elementAt(j);
						  for (int k=j+1; k<each.size(); k++)
						  {
								State q = (State)each.elementAt(k);
								boolean equiv = true;
								for (int l=0; (l<ta.size()) && equiv; l++)
								{
									 Character ct = (Character)ta.elementAt(l);
									 char t = ct.charValue();
									 int eq_p = inEquiv(indis0,m.transition(p,t));
									 int eq_q = inEquiv(indis0,m.transition(q,t));
									 if (eq_p != eq_q)
										  equiv = false;
								}
								if (equiv)
								{
									 Set s1 = setContaining(nEq,p);
									 Set s2 = setContaining(nEq,q);
									 if (s1 != s2)
									 {
										  s1.union(s2);
										  nEq.removeElement(s2);
									 }
								}
						  }
					 }
					 indis1.union(nEq);
				}
				else
				{
					 // the set has one element and so cannot be partitioned
					 // - just copy to subsequent equivalence class
					 indis1.addElement(each);
				}
		  }
		  return indis1;
    }

    /**
     * Return the set containing the given object, from the given set of sets.
     * @param all the set of sets to be searched
     * @param obj the object to be searched for
     * @return the set containing the given object
     */
    private Set setContaining(Set all,Object obj)
    {
		  for (int n=0; n<all.size(); n++)
		  {
				Set t = (Set)all.elementAt(n);
				if (t.contains(obj))
					 return t;
		  }
		  return null;
    }

    /**
     * Return an integer identifying the equivalence class of the given state.
     * @param equivs the set of all equivalence classes
     * @param s the state to be identified
     * @return an integer representing the equivalence class containing s
     */
    private int inEquiv(Set equivs, State s)
    {
		  for (int i=0; i<equivs.size(); i++)
		  {
				Set indis = (Set)equivs.elementAt(i);
				if (indis.contains(s))
					 return i+1;
		  }
		  return 0;
    }
}
