/* 
 * The Finite State Machine Explorer
 * Copyright (c) 1995-1996
 * Matt Chapman (csuoq@csv.warwick.ac.uk)
 * Dept. Computer Science
 * University of Warwick, Coventry, UK.
 */

package Machine;

import java.awt.*;

/**
 * A general-purpose Panner widget; allows two-dimensional scrolling of an
 * on-screen thumb
 * @version    7 March 1996
 * @author     Matt Chapman
 */
public class Panner extends Canvas 
{
    // dimensions of the thumb and panner
    private int wid, hei;
    private int tlx,tly;
    private int lheight,lwidth;
    private int theight,twidth;
    private int xdiff,ydiff;
    boolean in_drag = false;

    // customisable colours
    Color thumbColour, barColour, textColour;
    
    /**
     * Constructs a new panner widget.
     * @param w,h the screen width and height of the widget
     * @param lx,ly the logical width and height of the panner
     * @param tx,ty the size of the thumb in logical units
     */
    public Panner(int w,int h, int lx, int ly, int tx, int ty) 
    {
		  wid=w;
		  hei=h;
		  lwidth=lx;
		  lheight=ly;
		  twidth=tx;
		  theight=ty;
		  setValue(0,0);
		  resize(w+1,h+1);
		  thumbColour = Color.lightGray;
		  barColour = Color.lightGray.darker();
    }

    /**
     * Called when the "thumb" of the panner is dragged.
     * Must be overridden to give the panner some behavior.
     */
    public void motion() 
    {
		  ;
    }

    /**
     * Called when the "thumb" of the panner is released.
     * Must be overridden to give the panner some behavior.
     */
    public void release() 
    {
		  ;
    }
      
    /**
     * Sets the current value for the panner. The thumb will move to
     * reflect the new setting.
     * @param x,y the new settings for the panner
     */
    public void setValue(int x, int y) 
    {
		  tlx=x;
		  tly=y;
		  repaint();
    }

    public void msetSize(int tx, int ty) 
    {
		  twidth=tx;
		  theight=ty;
		  // check thumb has not moved off edges of panner
		  if (tlx + twidth > lwidth)
				tlx = lwidth - twidth;
		  if (tly + theight > lheight)
				tly = lheight - theight;
    }
    
    /**
     * Returns the current X value for the panner.
     * @return the X value
     */
    public int getXValue() 
    {
		  return tlx;
    }

    /**
     * Returns the current Y value for the panner.
     * @return the Y value
     */
    public int getYValue() 
    {
		  return tly;
    }

    /**
     * Sets the color for the panner's thumb. The "thumb" is the box that
     * the user can slide around.
     * @param c the new thumb colour
     */
    public void setThumbColour(Color c) 
    {
		  thumbColour = c;
		  repaint();
    }

    /**
     * Sets the color for the panner's bar. The "bar" is the rectangle
     * that the thumb slides around in.
     * @param c the new bar colour
     */
    public void setBarColour(Color c) 
    {
		  barColour = c;
		  repaint();
    }

    /**
     * Redraw the panner.
     */
    public void paint(Graphics g) 
    {
		  g.setColor(barColour);
		  g.fill3DRect(0, 0,size().width, size().height, false);
		  g.setColor(thumbColour);
		  g.fill3DRect(wid*tlx/lwidth, hei*tly/lheight, wid*twidth/lwidth,
							hei*theight/lheight, true);
    }

    void handleMouse(int x, int y) 
    {
		  tlx = lwidth*x/wid - xdiff;
		  tly = lheight*y/hei - ydiff;
		  if (tlx < 0)
				tlx = 0;
		  if (tly < 0)
				tly = 0;
		  if (tlx + twidth > lwidth)
				tlx = lwidth - twidth;
		  if (tly + theight > lheight)
				tly = lheight - theight;

		  paint(getGraphics());
    }
    
    public boolean mouseDown (Event e, int x, int y) 
    {
		  int ntlx = lwidth*x/wid;
		  int ntly = lheight*y/hei;
		  if ((ntlx < tlx) || (ntlx > tlx + twidth))
				return false;
		  if ((ntly < tly) || (ntly > tly + theight))
				return false;
		  xdiff = ntlx - tlx;
		  ydiff = ntly - tly;
		  in_drag = true;
		  return true;
    }

    public boolean mouseDrag (Event e, int x, int y) 
    {
		  if (in_drag)
		  {
				handleMouse(x,y);
				motion();
		  }
		  return true;
    }

    public boolean mouseUp (Event e, int x, int y) 
    {
		  release();
		  in_drag = false;
		  return true;
    }
}

