/* 
 * The Finite State Machine Explorer
 * Copyright (c) 1995-1996
 * Matt Chapman (csuoq@csv.warwick.ac.uk)
 * Dept. Computer Science
 * University of Warwick, Coventry, UK.
 */

package Machine;

import java.io.*;
import java.util.Date;

/**
 * A class to generate rendering of a machine in the PostScript language
 * @version    3 March 1996
 * @author     Matt Chapman
 */
public final class Postscript
{
    private static RandomAccessFile fd = null;
    private static PrintStream out = null;
    private int scaling;

    /**
     * Draw the machine using commands of the PostScript language
     * @param m the machine to be rendered
     * @param filename the filename to save the PostScript commands in
     * @param title a string to print at the top of the page
     * @param portrait if true the page output will portrait in orientation
     * otherwise it will be landscape
     */
    public void PSpaint(Logical m, String filename, String title,
								boolean portrait)
    {
		  // dimensions of the page
		  scaling = portrait ? 576 : 756;

		  // open file for output
        try { 
            out = new PrintStream(new FileOutputStream(filename));
        }
        catch (Exception e) {
            System.out.println("Could not save to \"" + filename + "\".");
            System.out.println("Exception: " + e);
            return;
        }

		  // generate the PostScript
		  out.println("%!PS-Adobe-2.0");
		  if (title.length() > 0)
				out.println("%%Title: "+title);
		  else
				out.println("%%Title: Postscript version of a finite automata");
		  out.println("%%Creator: The Finite State Machine Explorer");
		  Date d = new Date();
		  out.println("%%CreationDate: "+d);
		  if (portrait)
				out.println("%%Orientation: Portrait");
		  else
				out.println("%%Orientation: Landscape");
		  out.println("%%EndComments");

		  if (portrait)
				out.println("18 230 translate");
		  else
				out.println("90 rotate 18 -540 translate");

		  if (title.length() > 0)
		  {
				out.println("/Times-Bold findfont 16 scalefont setfont");
				if (portrait)
					 out.println("0 410 moveto");
				else
					 out.println("0 510 moveto");
				out.println("("+title+")");
				out.println("dup stringwidth pop "+scaling+" exch sub 2 div");
				out.println("0 rmoveto show newpath");

		  }
		  out.println("/Courier-Bold findfont "+fontSize(m)+
						  " scalefont setfont");

		  State any;
		  if (m.numStates() > 0)
				any = m.stateAt(0);
		  else
				any = new State();
		  float rad = any.radius*scaling;
		  float smrad = rad*0.7f;

		  // draw connections
		  out.println("% draw connections");
		  for (int i=0; i<m.numConnections(); i++) 
		  {
				Connection c = m.connectionAt(i);
				c.calculatePos();
				Coord f = view(c.getFrom().getPos());
				Coord t = view(c.getTo().getPos());

				out.println("0 0 1 setrgbcolor");
				// connection is to and from the same state
				if (c.getFrom() == c.getTo()) 
				{
					 // draw connecting arc
					 // (unwanted part of oval will be obscured by the state)
					 out.println((f.x+rad) + " " + (f.y-rad) + " " + rad +
									 " 0 360 arc stroke");
				}
				else { // connection is between distinct states
					 // draw connecting line
					 out.println(f.x+" " + f.y + " moveto " + t.x + " " + t.y + 
									 " lineto stroke");
				}
				// draw arrow head
				Coord tip = view(c.arrowtip);
				Coord e1 = view(c.arrowedge1);
				Coord e2 = view(c.arrowedge2);
				arrowhead(tip,e1,e2);

				// draw transition details
				Coord tl = view(c.getTopLeft());
				Coord br = view(c.getBotRight());
				out.println(tl.x+" "+br.y+" moveto");
				out.println("0 0 0 setrgbcolor");
				out.println("("+c.getTrans()+") show newpath");
		  }

		  // draw line to start state
		  State start = m.getStart();
		  if (start != null) 
		  {
				Coord st = view(start.getPos());
				out.println((st.x-3*rad) + " " + st.y + " moveto");
				out.println("0 0 1 setrgbcolor");
				out.println((st.x-rad) + " " + st.y + " lineto stroke");
				Coord tip = new Coord(st.x-rad,st.y);
				float arrowwidth = 0.005f * scaling;
				Coord e1 = new Coord(tip.x-rad,tip.y+arrowwidth);
				Coord e2 = new Coord(tip.x-rad,tip.y-arrowwidth);
				arrowhead(tip,e1,e2);
		  }

		  // draw states
		  out.println("% draw states");
		  for (int i=0; i<m.numStates(); i++) 
		  {
				State s = m.stateAt(i);
				Coord p = view(s.getPos());
				out.println(p.x + " " + p.y + " " + rad + " 0 360 arc");
				out.println("gsave 1 1 0 setrgbcolor fill");
				out.println("grestore 0 0 0 setrgbcolor stroke");
				if (s.isAccepting())
					 out.println(p.x + " " + p.y + " " + smrad +
									 " 0 360 arc stroke");
		  }

		  out.println("showpage");
		  out.println("% done");

		  try
		  {
				out.close();
		  }
		  catch (Exception e)
		  {
				return;
		  }
    }

    private Coord view(Coord A)
    {
		  Coord B = new Coord(A.x*scaling,(0.6f-A.y)*scaling);
		  return B;
    }

    private void arrowhead(Coord tip, Coord e1, Coord e2)
    {
		  out.println(tip.x+" "+tip.y+" moveto");
		  out.println(e1.x+" "+e1.y+" lineto");
		  out.println(e2.x+" "+e2.y+" lineto");
		  out.println("0 0 1 setrgbcolor");
		  out.println("closepath fill");
    }

    private int fontSize(Logical m) 
    {
		  if (m.numConnections() > 0) 
		  {
				Connection c = m.connectionAt(0);
				c.calculatePos();
				Coord tl = view(c.getTopLeft());
				Coord br = view(c.getBotRight());
				float h = tl.y-br.y;
				// calculate required font size;
				int calc = (int)((h*2*1031)/1609.0f+0.5f);
				return calc;
		  }
		  return 1;
    }
}
