/* 
 * The Finite State Machine Explorer
 * Copyright (c) 1995-1996
 * Matt Chapman (csuoq@csv.warwick.ac.uk)
 * Dept. Computer Science
 * University of Warwick, Coventry, UK.
 */

package Machine;

import java.util.Vector;

/**
 * Implements a general-purpose Set of objects.
 * @version    10 Feb 1996
 * @author     Matt Chapman
 */
public final class Set 
{

    // a Vector is used to store the set
    private Vector set_v;

    /**
     * Creates a new, empty Set.
     */
    public Set() 
    {
		  set_v = new Vector();
    }

    /**
     * Adds an element to the set. If the element is already in the set, this
     * will have no effect.
     * @param obj the object to be added
     */
    public void addElement(Object obj) 
    {
		  if (!set_v.contains(obj)) 
		  {
				set_v.addElement(obj);
		  }
    }

    /**
     * Remove an element from the set.
     * @param obj the object to be removed
     */
    public void removeElement(Object obj) 
    {
		  set_v.removeElement(obj);
    }

    /**
     * Test membership of the set.
     * @param obj the object of interest
     * @return true if the object is in the set, false otherwise
     */
    public boolean contains(Object obj) 
    {
		  return set_v.contains(obj);
    }

    /**
     * Perform the union operation with another set.
     * @param s the set to be merged with
     */
    public void union(Set s) 
    {
		  for (int n=0; n<s.size(); n++) 
		  {
				this.addElement(s.elementAt(n));
		  }
    }

    /**
     * Perform the intersection operation with another set.
     * @param s the set to be intersected
     * @return the intersection set
     */
    public Set intersection(Set s) 
    {
		  Set in = new Set();
		  for (int i=0; i<this.size(); i++) 
		  {
				Object o = this.elementAt(i);
				if (s.contains(o)) 
				{
					 in.addElement(o);
				}
		  }
		  return in;
    }

    /**
     * Return the cardinality - the number of elements in the set.
     * @return the cardinality of the set
     */
    public int size() 
    {
		  return set_v.size();
    }

    /**
     * Return an element of the set. This should only be used to iterate over
     * elements in the set - the elements will be returned in no specific
     * order.
     * @param n the number of the element required - from 0 to less than the
     * cardinality of the set.
     * @return an element of the set
     */
    public Object elementAt(int n) 
    {
		  return set_v.elementAt(n);
    }

    /**
     * Make a copy of the current set.
     * @return an object containing the same elements as the set
     */
    public Object clone() 
    {
		  Set s = new Set();
		  for (int i=0; i<this.size(); i++) 
		  {
				s.addElement(this.elementAt(i));
		  }
		  return s;
    }

    /**
     * Determines whether two sets contain the same elements.
     * @param s the set to compare this one with
     * @return true if the sets contain the same elements, else false
     */
    public boolean equals(Set s)
    {
		  if (s.size() != this.size())
				return false;

		  boolean same = true;
		  for (int i=0; (i<this.size()) && same; i++) 
		  {
				Object o = this.elementAt(i);
				boolean found = false;
				for (int j=0; (j<s.size()) && !found; j++) 
				{
					 Object p = s.elementAt(j);
					 if (p.equals(o))
						  found = true;
				}
				same = found;
		  }
		  return same;
    }

    /**
     * Generates the powerset of the set.
     * @returns the powerset
     */
    public Set powerSet() 
    {
		  Set Z = this;
		  Set X = new Set();
		  Set s = new Set();
		  X.addElement(s);
		  Set Y = new Set();
		  Y.addElement(s);
		  for (int i=0; i<Z.size(); i++) 
		  {
				Object z = Z.elementAt(i);
				for (int j=0; j<X.size(); j++) 
				{
					 Set x = (Set)X.elementAt(j);
					 Set t = new Set();
					 t.addElement(z);
					 t.union(x);
					 Y.addElement(t);
				}
				X = (Set)Y.clone();
		  }	
		  return Y;
    }

    /**
     * Return a String represention of the set, by concatenating the String
     * form of all the elements of the set, in no particular order.
     * @return the String representation
     */
    public String toString() 
    {
		  String s = new String("{");
		  for (int n=0; n<size(); n++) 
		  {
				s = s + " " + elementAt(n).toString();
		  }
		  s = s + "}";
		  return s;
    }
}
