/* 
 * The Finite State Machine Explorer
 * Copyright (c) 1995-1996
 * Matt Chapman (csuoq@csv.warwick.ac.uk)
 * Dept. Computer Science
 * University of Warwick, Coventry, UK.
 */

package Machine;

import java.awt.*;
import java.util.Date;

/**
 * Simulation for a finite state machine, given a particular input.
 * @version    9 Feb 1996
 * @author     Matt Chapman
 */
public final class Simulate implements Runnable 
{

    // delay in milliseconds between steps of the simulation
    private final int SLOW = 4000;
    private final int MEDIUM = 2000;
    private final int FAST = 1000;
    private final int VFAST = 200;

    // AWT components
    Label msg_label;
    Logical m;
    String input;
    World screen;
    Graphics g;
    MainWidget mainW;
    Button simbutton;

    int delay=0;
    boolean instant = false;

    // the simulation is to run in a separate thread
    Thread sthread = null;

    // change the messase string
    public void msg(String txt) 
    {
		  msg_label.setText(txt);
    }

    /**
     * Prepare for simulation by setting mesage label.
     * @param msg a label component for messages
     */
    public Simulate(Label msg) 
    {
		  msg_label = msg;
    }

    /**
     * Validates an input string, by checking for zero length and invalid
     * characters.
     * @param input the input string
     * @return true if valid, else false
     */
    public boolean validInput(String input) 
    {
		  if (input.length() == 0) 
		  {
				msg("Enter an input string first");
				return false;
		  }
		  for (int pos=0; pos<input.length(); pos++) 
		  {
				char k = input.charAt(pos);
				if (!((k>='a' && k<='z') || (k>='A' && k<='Z')
						|| (k>='0' && k<='9'))) 
				{
					 msg("The input string contains invalid letters");
					 return false;
				}
		  }
		  return true;
    }

    /**
     * Validates a given machine, by checking for absent starting state.
     * @param m the machine of interest
     * @return true if valid, else false
     */
    public boolean validMachine(Logical m) 
    {
		  if (m.getStart() == null) 
		  {
				msg("The machine must have a starting state");
				return false;
		  }
		  return true;
	 }

    /**
     * Start the simulation thread.
     * @param inp the input string
     * @param gr the graphics context, for drawing operations
     * @param mw the mainwidget
     * @param sc the world class
     * @param mc the machine to be simulated
     * @param sb a button for aborting the simulation
     * @param speedbutt a choice component for changing the simulation speed
     */
    public void runSim(String inp,Graphics gr,MainWidget mw,
							  World sc,Logical mc,Button sb,Choice speedbutt) 
	 {
		  m=mc;
		  g=gr;
		  input=inp;
		  mainW=mw;
		  screen=sc;
		  simbutton=sb;
		  int sp = speedbutt.getSelectedIndex();
		  if (sp == 0)
				delay = SLOW;
		  else if (sp == 1)
				delay = MEDIUM;
		  else if (sp == 2)
				delay = FAST;
		  else if (sp == 3)
				delay = VFAST;
		  else if (sp == 4)
				instant = true;
		  simbutton.setLabel("Stop");
		  start();
    }

    public void start() 
    {
		  // Create a new thread
		  // "this" refers to the current class which is a Runnable
		  if (sthread == null)	
		  {
				sthread = new Thread(this);
				//one.setPriority(Thread.MIN_PRIORITY);
				sthread.start();
		  }
    }

    // wait for the given period
    void pause(int delay) 
    {
		  for (int i=0; i<delay/100; i++) 
		  {
				try 
				{
					 sthread.sleep(100);
				}
				catch(InterruptedException e) {}
				if (sthread == null) return;
		  }
    }

    // called when the thread is created
    public void run() 
    {
		  Set cstates = new Set();
		  Set newstates;
		  cstates.addElement(m.getStart());
		  if (!instant)
				msg("Simulation of machine with input: "+input+"...");

		  for (int pos=0; pos<input.length() && (sthread!=null); pos++) 
		  {
				if (!instant) 
				{
					 screen.hilight = cstates;
					 screen.simstring = "Remaining input: "+input.substring(pos);
					 mainW.paint(g);
					 pause(delay);
				}
				char trans = input.charAt(pos);
				newstates = new Set();
				// traverse sets of current states
				for (int n=0; n<cstates.size(); n++) 
				{
					 State cs = (State)cstates.elementAt(n);
					 newstates.union(m.transitionSet(cs,trans));
					 if (!instant) 
					 {
						  msg("Input character: "+trans);	    
					 }
				}
				cstates = new Set();
				cstates.union(newstates);

		  }

		  if (sthread == null) 
		  {
				msg("Simulation cancelled");
				return;
		  }

		  if (!instant) 
		  {
				screen.hilight = cstates;
				mainW.paint(g);
		  }
		  // end of input - check for accepting states

		  boolean accept = false;
		  for (int n=0; n<cstates.size() && !accept; n++) 
		  {
				State cs = (State)cstates.elementAt(n);
				if (cs.isAccepting())
					 accept = true;
		  }

		  if (accept) { // accept
				msg("Input ACCEPTED");
		  }
		  else 
		  {
				msg("Input REJECTED");
		  }

		  if (!instant) 
		  {
				pause(delay);
		  }
		  stop();
    }

    public void stop()    
    {
		  // kill thread
		  sthread = null;
		  simbutton.setLabel("Simulate");
		  screen.hilight = null;
		  screen.simstring = null;
		  mainW.paint(g);
    }

}
