/* 
 * The Finite State Machine Explorer
 * Copyright (c) 1995-1996
 * Matt Chapman (csuoq@csv.warwick.ac.uk)
 * Dept. Computer Science
 * University of Warwick, Coventry, UK.
 */

package Machine;

import java.awt.*;

/**
 * A general-purpose Slider widget; allows one-dimensional scrolling of an
 * on-screen thumb
 * @version    14 Jan 1996
 * @author     Matt Chapman
 */
public class Slider extends Canvas 
{
    private final static int THUMBSIZE = 14;
    private final static int BUFFER = 2;

    private final static int TEXTHEIGHT = 18;
    private final static int TEXTBUFFER = 3;
    
    private final static int DEFAULTWIDTH = 100;
    private final static int DEFAULTHEIGHT = 15;

    private final static int MINWIDTH = 2 * (THUMBSIZE + BUFFER + 1);
    private final static int MINHEIGHT = 2 * (BUFFER + 1);

    private final static int DEFAULTMIN = 1;
    private final static int DEFAULTMAX = 100;    
    
    int min, max, value, pixel;
    int pixelMin, pixelMax;
    Color backgroundColor, thumbColor, barColor, slashColor, textColor;
    Font font;
    
    /**
     * Constructs a slider.
     * @param container The container for this slider.
     */
    public Slider() 
    {
		  min = DEFAULTMIN;
		  max = DEFAULTMAX;
		  resize(DEFAULTWIDTH, DEFAULTHEIGHT + TEXTHEIGHT);
		  font = new Font("TimesRoman", Font.PLAIN, 12);
		  backgroundColor = Color.lightGray;
		  thumbColor = Color.lightGray;
		  barColor = Color.lightGray.darker();
		  slashColor = Color.black;
		  textColor = Color.black;
		  SetValue(1);
    }

    /**
     * Called when the "thumb" of the slider is dragged.
     * Must be overridden to give the slider some behavior.
     */
    public void motion()
    {
		  ;
    }

    /**
     * Called when the "thumb" of the slider is released.
     * Must be overridden to give the slider some behavior.
     */
    public void release()
    {
		  ;
    }
    
    /**
     * Sets the maximum value for the slider.
     * @param num The new maximum.
     */
    public void SetMaximum(int num) 
    {
		  max = num;
		  if (max < min) 
		  {
				int t = min;
				min = max;
				max = t;
		  }
		  SetValue(value);
    }
    
    /**
     * Sets the minimum value for the slider.
     * @param num The new minimum.
     */
    public void SetMinimum(int num) 
    {
		  min = num;
		  if (max < min) 
		  {
				int t = min;
				min = max;
				max = t;
		  }
		  SetValue(value);
    }
    
    /**
     * Sets the current value for the slider. The thumb will move to
     * reflect the new setting.
     * @param num The new setting for the slider.
	  */
    public void SetValue(int num) 
    {
		  value = num;
	
		  if (value < min)
				value = min;
		  else if (value > max)
				value = max;
	
		  if (value != min)
				pixel = (int)(Math.round(Math.abs((double)(value - min) /
															 (double)(max - min)) *
												 (double)(pixelMax - pixelMin)) +
								  pixelMin);
		  else
				pixel = pixelMin;

		  repaint();
    }
    
    /**
     * Sets the height of the slider. This is the height of the entire
     * slider canvas, including space reserved for displaying the
     * current value.
     * @param num The new height.
     */
    public void SetHeight (int num) 
    {
		  if (num < MINHEIGHT + TEXTHEIGHT)
				num = MINHEIGHT + TEXTHEIGHT;
		  resize(size().width, num);
		  repaint();
    }
    
    /**
     * Sets the width of the slider. This is the width of the actual
     * slider box.
     * @param num The new width.
     */
    public void SetWidth(int num) 
    {
		  if (num < MINWIDTH)
				num = MINWIDTH;
		  resize(num, size().height);
		  repaint();	
    }
    
    /**
     * Returns the current value for the slider.
     * @return The current value for the slider.
     */
    public int GetValue() 
    {
		  return value;
    }

    /**
     * Sets the background color for the slider. The background is the
     * area outside of the bar.
     * @param color The new background color.
     */
    public void SetBackgroundColor(Color color) 
    {
		  backgroundColor = color;
		  repaint();
    }

    /**
     * Sets the color for the slider's thumb. The thumb is the box that
     * the user can slide back and forth.
     * @param color The new thumb color.
     */
    public void SetThumbColor(Color color) 
    {
		  thumbColor = color;
		  repaint();
    }

    /**
     * Sets the color for the slider's bar. The bar is the rectangle
     * that the thumb slides around in.
     * @param color The new bar color.
     */
    public void SetBarColor(Color color) 
    {
		  barColor = color;
		  repaint();
    }

    /**
     * Sets the slash color for the slider. The slash is the
     * vertical line on the thumb.
     * @param color The new slash color.
     */
    public void SetSlashColor(Color color) 
    {
		  slashColor = color;
		  repaint();
    }

    /**
     * Sets the color for the slider's text.
     * @param color The new text color.
     */
    public void SetTextColor(Color color) 
    {
		  textColor = color;
		  repaint();
    }

    /**
     * Sets the font for the slider's text.
     * @param font The new font.
     */
    public void SetFont(Font font) 
    {
		  font = font;
		  repaint();
    }
    
    public void paint(Graphics g) 
    {
		  int width = size().width;	
		  int height = size().height;

		  g.setColor(backgroundColor);
		  g.fillRect(0, 0, width, TEXTHEIGHT);

		  g.setColor(barColor);
		  g.fill3DRect(0, TEXTHEIGHT,
							width, height - TEXTHEIGHT, false);

		  g.setColor(thumbColor);	
		  g.fill3DRect(pixel - THUMBSIZE, TEXTHEIGHT + BUFFER,
							THUMBSIZE * 2 + 1, height - 2 * BUFFER - TEXTHEIGHT,
							true);
	
		  g.setColor(slashColor);
		  g.drawLine(pixel, TEXTHEIGHT + BUFFER + 1,
						 pixel, height - 2 * BUFFER);

		  g.setColor(textColor);
		  g.setFont(font);		
		  String str = String.valueOf(value);
		  g.drawString(str, pixel -
							(int)(getFontMetrics(font).stringWidth(str) / 2),
							TEXTHEIGHT - TEXTBUFFER);
    }

    void handleMouse(int x) 
    {
		  double percent;
		  int width = size().width;
		  pixel = Math.max(x, pixelMin);
		  pixel = Math.min(pixel, pixelMax);

		  if (pixel != pixelMin)
				percent = (((double)pixel - pixelMin) /
							  (pixelMax - pixelMin));
		  else
				percent = 0;
	
		  value = (int)(Math.round(percent * (double)(max - min))) + min;
	
		  paint(getGraphics());
    }
    
    public boolean mouseDown(Event e, int x, int y) 
    {
		  handleMouse(x);
		  motion();
		  return true;
    }

    public boolean mouseDrag(Event e, int x, int y) 
    {
		  handleMouse(x);
		  motion();	
		  return true;
    }

    public boolean mouseUp(Event e, int x, int y) 
    {
		  handleMouse(x);
		  release();
		  return true;
    }

    public void reshape(int x, int y, int width, int height) 
    {
		  super.reshape(x, y, width, height);
		  pixelMin = THUMBSIZE + BUFFER;
		  pixelMax = width - THUMBSIZE - BUFFER - 1;
		  if (value != min)
				pixel = (int)(Math.round(Math.abs((double)(value - min) /
															 (double)(max - min)) *
												 (double)(pixelMax - pixelMin)) +
								  pixelMin);
		  else
				pixel = pixelMin;
    }
}

