/* 
 * The Finite State Machine Explorer
 * Copyright (c) 1995-1996
 * Matt Chapman (csuoq@csv.warwick.ac.uk)
 * Dept. Computer Science
 * University of Warwick, Coventry, UK.
 */

package Machine;

import java.awt.*;
import java.util.*;
import java.applet.Applet;

/**
 * Represents the real world interpretation of a finite state machine.
 * @version    21 Jan 1996
 * @author     Matt Chapman
 */
public final class World extends java.applet.Applet 
{

    // the current scale factor
    private float scaling = 0.5f;
    private final int sf = 570;
    private boolean scaling_changed = true;

    // the topleft of the viewport
    private Coord topleft = new Coord();

    // the left and right offset of the window - the portion of the object
    // space to be viewed
    private float WxL = 0.0f, WyB = 0.0f;

    // the font for the transition characters
    private Font tfont = null;
    private Font dfont = new Font("Courier", Font.BOLD, 12);

    public Set hilight = null;
    public Image buttons;
    public String simstring = null;

    /**
     * Set the window position - the portion of the machine to be viewed.
     * @param xoff the X-offset of the window
     * @param yoff the Y-offset of the window
     */
    public void setWindow(int xoff, int yoff) 
    {
		  WxL = (float)xoff;
		  WyB = (float)yoff;
    }

    /**
     * Set the scaling used to display the machine.
     * @param s the new scale factor
     */
    public void setScaling(float s) 
    {
		  scaling = s;
		  scaling_changed = true;
    }

    /**
     * The viewing pipeline. Converts a coordinate in the normalized object
     * space to a physical screen coordinate, using the current scaling and
     * window position information.
     */
    public Point viewPipeline(Coord A) 
    {
		  Point B = new Point((int)((sf*A.x - WxL)/scaling),
									 (int)((sf*A.y - WyB)/scaling));
		  return B;
    }

    /**
     * Reverse of the viewing pipeline. Converts a physical screen coordinate
     * into the corresponding coordinate in the normalized object space,
     * using the current scaling and window position information.
     */
    public Coord reversePipeline(Point B) 
    {
		  Coord A = new Coord( (((float)B.x)*scaling + WxL) / sf ,
									  (((float)B.y)*scaling + WyB) / sf );
		  return A;
    }

    public float viewSc(float a)
    {
		  return sf*a/scaling;
    }

    /**
     * Find the first state in the given machine (if there is one) at the
     * given physical screen coordinate.
     * @param m the machine of interest 
     * @param p the coordinate of interest
     * @returns the state at the given position, or null if there is no such
     * state
     */
    public State getState(Logical m, Point p) 
    {
		  Coord u = reversePipeline(p);
		  for (int i=0; i<m.numStates(); i++) 
		  {
				State s = m.stateAt(i);
				Coord t = s.getPos();
				if ((t.x-u.x)*(t.x-u.x) + (t.y-u.y)*(t.y-u.y) <= s.radius*s.radius)
					 return s;
		  }
		  return null; // no state found
    }

    /**
     * Determines whether there is room for a state at the given physical
     * screen coordinate.
     * @param m the machine of interest 
     * @param p the coordinate of interest
     * @returns true if there is room, false otherwise
     * state
     */
    public boolean roomFor(Logical m, Point p) 
    {
		  Coord u = reversePipeline(p);
		  for (int i=0; i<m.numStates(); i++) 
		  {
				State s = m.stateAt(i);
				Coord t = s.getPos();
				if ((t.x-u.x)*(t.x-u.x) + (t.y-u.y)*(t.y-u.y)
					 <= 4*s.radius*s.radius)
					 return false;
		  }
		  return true;
    }

    /**
     * Determines whether there is room for a state at the given physical
     * screen coordinate, ignoring a particular state.
     * @param m the machine of interest 
     * @param p the coordinate of interest
     * @param ignore the state to igore
     * @returns true if there is room, false otherwise
     */
    public boolean roomFor(Logical m, Point p, State ignore) 
    {
		  Coord u = reversePipeline(p);
		  for (int i=0; i<m.numStates(); i++) 
		  {
				State s = m.stateAt(i);
				if (s != ignore) 
				{
					 Coord t = s.getPos();
					 if ((t.x-u.x)*(t.x-u.x) + (t.y-u.y)*(t.y-u.y)
						  <= 4*s.radius*s.radius)
						  return false;
				}
		  }
		  return true;
    }

    /**
     * Find the first connection at the given coordinate in the normalized
     * object space.
     * @param m the machine of interest 
     * @param p the coordinate of interest
     * @returns the connection at the given position, or null if there is no
     * such state
     */
    public Connection getConnection(Logical m, Point p) 
    {
		  Coord d = reversePipeline(p);

		  for (int i=0; i<m.numConnections(); i++) 
		  {
				Connection c = m.connectionAt(i);
				Coord tl = c.getTopLeft();
				Coord br = c.getBotRight();
				if ((d.x >= tl.x) && (d.y >= tl.y)
					 && (d.x <= br.x) && (d.y <= br.y))
					 return c;
		  }
		  return null; // no connection found
    }

    void determineFont(Logical m) 
    {
		  if (m.numConnections() > 0) 
		  {
				Connection c = m.connectionAt(0);
				c.calculatePos();
				Point tl = viewPipeline(c.getTopLeft());
				Point br = viewPipeline(c.getBotRight());
				int h = br.y-tl.y;
				// calculate required font size;
				int calc = (int)((float)(h*2062)/1609.0f+0.5f);
				tfont = new Font("Courier", Font.BOLD, calc);
		  }
    }

    /* draws a triangle through the given 3 points */
    private void arrowhead(Point tip, Point e1, Point e2, Graphics g) 
    {
		  int xpns[] = new int[3];
		  int ypns[] = new int[3];
		  xpns[0] = tip.x; ypns[0] = tip.y;
		  xpns[1] = e1.x;  ypns[1] = e1.y;
		  xpns[2] = e2.x;  ypns[2] = e2.y;
		  g.fillPolygon(xpns,ypns,3);
    }

    /**
     * Paint the entire machine representation to the given graphics object.
     * @param m the machine of interest 
     * @param g the graphics object to paint to
     */
    public synchronized void paintMachine(Logical m, Graphics g, int wid,
														int hei,int mode) 
    {
		  State any;
		  if (m.numStates() > 0)
				any = m.stateAt(0);
		  else
				any = new State();
		  int dia = (int)viewSc(2*any.radius);
		  int rad = dia/2;

		  // smaller sizes used to indicate accepting states
		  int smdia = (int)viewSc(any.radius*2.0f*0.7f);
		  int smrad = smdia/2;

		  // recalculate font size only if scaling has changed
		  if ((scaling_changed) || (tfont == null)) 
		  {
				determineFont(m);
				scaling_changed = false;
		  }

		  if (tfont != null) g.setFont(tfont);
		  g.setColor(Color.lightGray);
		  g.fillRect(0,0,wid,hei);
		  g.setColor(Color.red);
		  g.draw3DRect(0, 0, wid-1, hei-1, true);

		  // draw connections
		  for (int i=0; i<m.numConnections(); i++) 
		  {
				Connection c = m.connectionAt(i);
				c.calculatePos();
				Point f = viewPipeline(c.getFrom().getPos());
				Point t = viewPipeline(c.getTo().getPos());
				g.setColor(Color.blue);

				// connection is to and from the same state
				if (c.getFrom() == c.getTo()) 
				{
					 // draw connecting arc
					 // (unwanted part of oval will be obscured by the state)
					 g.drawOval((int)f.x,f.y,dia,dia);
				}
				else { // connection is between distinct states
					 // draw connecting line
					 g.drawLine(f.x, f.y, t.x, t.y);
				}
		
				// draw arrow head
				Point tip = viewPipeline(c.arrowtip);
				Point e1 = viewPipeline(c.arrowedge1);
				Point e2 = viewPipeline(c.arrowedge2);
				arrowhead(tip,e1,e2,g);

				// draw transition details
				Point tl = viewPipeline(c.getTopLeft());
				Point br = viewPipeline(c.getBotRight());
				g.setColor(Color.black);
				g.drawString(c.getTrans(),tl.x+2,br.y);
		  }

		  // draw line to start state
		  State start = m.getStart();
		  if (start != null) 
		  {
				Point st = viewPipeline(start.getPos());
				g.setColor(Color.blue);
				g.drawLine(st.x-rad-dia, st.y, st.x-rad, st.y);
				Point tip = new Point(st.x-rad,st.y);
				int arrowwidth = (int)viewSc(0.005f);
				Point e1 = new Point(tip.x-rad,tip.y+arrowwidth);
				Point e2 = new Point(tip.x-rad,tip.y-arrowwidth);
				arrowhead(tip,e1,e2,g);
		  }

		  // draw states
		  for (int i=0; i<m.numStates(); i++) 
		  {
				State s = m.stateAt(i);
				Point p = viewPipeline(s.getPos());

				if ((hilight != null) && hilight.contains(s))
					 g.setColor(Color.green);
				else
					 g.setColor(Color.yellow);

				g.fillOval(p.x - rad, p.y - rad, dia, dia);
				g.setColor(Color.black);
				g.drawOval(p.x - rad, p.y - rad, dia, dia);
				if (s.isAccepting())
					 g.drawOval(p.x - smrad, p.y - smrad, smdia, smdia);
		  }

		  // draw buttons
		  int bw = buttons.getWidth(this);
		  int bh = buttons.getHeight(this);
		  int eh = (int)((bh-1)/6);
		  g.setColor(Color.white);
		  g.fillRect(wid - bw,0,bw,bh);
		  g.setColor(Color.gray);
		  g.fillRect(wid - bw,eh*(mode-1),bw,eh);
		  g.drawImage(buttons,wid - bw,0,this);

		  // print simulation string, if defined
		  if (simstring != null)
				g.drawString(simstring,wid-250,hei-15);
    }
}
